package info.textgrid.lab.collatex.process;

import info.textgrid.lab.collatex.CollateXPlugin;
import info.textgrid.lab.collatex.model.CollationResults;
import info.textgrid.lab.collatex.model.CollationSet;
import info.textgrid.lab.collatex.model.CollationSetWitness;
import info.textgrid.lab.collatex.model.equivalences.Equivalences;
import info.textgrid.lab.collatex.model.text.BasicWitness;
import info.textgrid.lab.core.model.CrudServiceException;
import info.textgrid.lab.core.model.TextGridObject;
import info.textgrid.lab.core.swtutils.AdapterUtils;
import info.textgrid.lab.core.swtutils.AdapterUtils.AdapterNotFoundException;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collection;

import javax.xml.bind.JAXB;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.IJobChangeListener;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.osgi.util.NLS;
import org.eclipse.ui.statushandlers.StatusManager;
import org.neo4j.graphdb.Transaction;

import com.google.common.collect.Collections2;
import com.google.common.io.CharStreams;

import eu.interedition.collatex.CollationAlgorithm;
import eu.interedition.collatex.CollationAlgorithmFactory;
import eu.interedition.collatex.graph.GraphFactory;
import eu.interedition.collatex.graph.VariantGraph;
import eu.interedition.collatex.matching.EqualityTokenComparator;
import eu.interedition.collatex.simple.WhitespaceTokenizer;

/**
 * Performs the actual collation including all the long-running operations that
 * are required for this.
 * 
 * Clients instantiate this class with the model information required to perform
 * the collation and schedule it when the user wants to perform the collation.
 * They should {@linkplain #addJobChangeListener(IJobChangeListener) register} a
 * {@link IJobChangeListener} and retrieve the status and collation results when
 * the collation has finished. The job reports progress and supports
 * cancellation, but especially for large collation sets it may take a while
 * until it reacts to a cancel operation.
 */
public class CollateJob extends Job {

	private final CollationResults collationResults;

	public CollateJob(final String name, final CollationSet collationSet) {
		super(name);
		collationResults = new CollationResults();
		collationResults.setCollationSet(collationSet);
		setUser(true);
	}

	public CollationResults getCollationResults() {
		return collationResults;
	}

	@Override
	protected IStatus run(final IProgressMonitor monitor) {
		final Collection<CollationSetWitness> witnesses = Collections2
				.filter(
						collationResults.getCollationSet().getWitnesses(),
						CollationSetWitness.IS_ACTIVE);
		final SubMonitor progress = SubMonitor.convert(monitor,
				10 + 10 * witnesses.size());

		EqualityTokenComparator comparator;

		if (getCollationSet().getRegularisation() != null) {
			final TextGridObject regObject = TextGridObject
					.getInstanceOffline(getCollationSet().getRegularisation());
			final IFile file = AdapterUtils.getAdapter(regObject, IFile.class);
			if (file != null)
				try {
					final Equivalences equivalences = JAXB.unmarshal(
							file.getContents(), Equivalences.class);
					getCollationResults().setEquivalences(equivalences);

				} catch (final CoreException e) {
					StatusManager.getManager().handle(e,
							CollateXPlugin.PLUGIN_ID);
				}
		}

		comparator = getCollationResults().getComparator();
		final CollationAlgorithm algorithm = CollationAlgorithmFactory
				.dekker(comparator);

		try {
			// FIXME this creates a variant graph using the Simple data model --
			// we should switch to our own stuff when everything has been
			// implemented
			// FIXME however SimpleToken seems to be used in a bunch of classes
			// from CollateX
			final GraphFactory graphFactory = GraphFactory.create();
			progress.worked(1);
			final Transaction newTx = graphFactory.getDatabase().beginTx();
			try {
				collationResults
				.setVariantGraph(graphFactory.newVariantGraph());
				newTx.success();
			} finally {
				newTx.finish();
			}
			progress.worked(4);

			for (final CollationSetWitness collationSetWitness : witnesses) {
				if (progress.isCanceled())
					return Status.CANCEL_STATUS;
				progress.subTask(NLS.bind(Messages.CollateJob_Reading_x, collationSetWitness));
				final BasicWitness witness = readWitness(collationSetWitness);
				progress.subTask(NLS.bind(Messages.CollateJob_Collating_x, collationSetWitness));
				progress.worked(1);
				final Transaction transaction = getVariantGraph().getDatabase()
						.beginTx();
				try {
					algorithm.collate(getVariantGraph(), witness);
					transaction.success();
				} finally {
					transaction.finish();
				}
				progress.worked(9);

			}

			progress.done();

		} catch (final IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return Status.OK_STATUS;
	}

	public CollationSet getCollationSet() {
		return collationResults.getCollationSet();
	}

	/**
	 * Returns the {@link VariantGraph} created by the collation process, or
	 * <code>null</code> if the job has not started yet.
	 * <strong>Warning:</strong> It is possible to retrieve a
	 * {@link VariantGraph} while the job is still running, but results may be
	 * undefined. I.e. you don't want to do that.
	 */
	public VariantGraph getVariantGraph() {
		return collationResults.getVariantGraph();
	}

	private BasicWitness readWitness(final CollationSetWitness witness)
			throws IOException {
		try {
			final TextGridObject object = TextGridObject.getInstance(
					witness.getUri(), false);
			final IFile file = AdapterUtils.getAdapterChecked(object,
					IFile.class);
			final InputStreamReader contents = new InputStreamReader(
					file.getContents(true), "UTF-8"); //$NON-NLS-1$
			try {
				final String witnessString = CharStreams.toString(contents);
				final BasicWitness result = new BasicWitness(witness,
						witnessString,
						new WhitespaceTokenizer());
				return result;
			} finally {
				contents.close();
			}
		} catch (final CrudServiceException e) {
			throw new IOException(NLS.bind(Messages.CollateJob_Error_reading_x, witness), e);
		} catch (final AdapterNotFoundException e) {
			throw new IOException(NLS.bind(Messages.CollateJob_Error_reading_x_content,
					witness), e);
		} catch (final CoreException e) {
			throw new IOException(NLS.bind(Messages.CollateJob_Error_reading_x, witness), e);
		}
	}

}
